/*
 * TADifference.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 23-set-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Vector;

import xrel.XreError;
import xrel.parser.SpecialNames;
import xrel.parser.SymTable;

/**
 * TODO insert comment here
 * @author Fabrizio Bisi
 */
public class TADifference extends TA implements SpecialNames {
	private static final String nullAutomatonName = "null";
	private TA fstAutomaton, sndAutomaton;
	
	public static String getDiffName(TA a1, TA a2) {
		return a1 + "\\" + a2;
	}
//	// aList is a list of automata
//	public static String getDiffName(TA a1, HashSet aList) {
//		return a1 + "\\" + aList;
//	}
	// aList is a list of automata names
	public static String getDiffName(String a1Name, HashSet aList) {
		return a1Name + "\\" + aList;
	}
	
	public TADifference(TA A1, TA A2, SymTable st, boolean dbg, PrintStream psOut) {
		super (getDiffName(A1,A2),st,dbg,psOut);
		fstAutomaton = A1;
		sndAutomaton = A2;
	}
	
	public void build() {
		// first of all ensure that source automata are full built
		fstAutomaton.update(STEP_BUILD | /* STEP_NO_EPS_TRANS | */ 
												STEP_NO_UNREACH_STATES | STEP_COMPLETE_AUT | 
												STEP_NO_UNMATCHED_STATES);
		sndAutomaton.update(STEP_BUILD | /* STEP_NO_EPS_TRANS | */ 
												STEP_NO_UNREACH_STATES | STEP_COMPLETE_AUT | 
												STEP_NO_UNMATCHED_STATES);
		// Clean states
		states.clear();
		initialStates.clear();
		finalStates.clear();
		transitions.clear();
		epsilonTransitions.clear();
		// sList is the set of initial states of the second automaton
		HashSet sList = new HashSet();
		sList.addAll(sndAutomaton.initialStates);
		// check if sList is final
		boolean slFinal = isDiffFinal(sList);
		// now add initial states
		// Initial states are as much as those of first automaton
		for (int i=0; i<fstAutomaton.initialStates.size(); i++) {
			TAState s1 = (TAState)fstAutomaton.initialStates.elementAt(i);
			boolean s1Final = s1.isFinal();
			// creates only if it's not a null state
			//if (!s1Final || !slFinal) {
				// create the subset state that collects states of second automaton
				TADifferenceState s = new TADifferenceState(s1,sList);
				states.add(s);
				initialStates.add(s);
//				s.setInitial(true);
				/* final only if the first is final and the second is not final
				 * (remember we compute the difference: A\B recognizes the
				 * empty string if and only if A recognizes the empty string
				 * but B don't recognize it) */
				if (s1Final && !slFinal) {
					s.setFinal(true);
					finalStates.add(s);
				}
			//}
		}
		// Then iterate adding transitions and the other states
		for (int i=0; i<states.size(); i++) {
			TADifferenceState s = (TADifferenceState)states.elementAt(i);
			processState(fstAutomaton,sndAutomaton,s);
		}
	}

	private void processState(TA A1, TA A2,TADifferenceState s) {
		// build the list of transitions exiting from the difference states
		Vector slTransitions = new Vector();
		Iterator it = s.getStateList().iterator();
		while (it.hasNext()) {
			TAState slEl = (TAState)it.next();
			slTransitions.addAll(slEl.getTransitions());
		}
		
		// Label normalization
		Vector srcTransitions =
		 normalizeLabels(s.getState1().getTransitions(),slTransitions);

		// now loops through transitions
		for (int i=0; i< srcTransitions.size(); i++) {
			TATransition t1 =
			  (TATransition)srcTransitions.elementAt(i);
			// build the list of transitions label-matching with t1
			Vector tList = new Vector();
			String label=t1.label;
			for (int j=0; j < slTransitions.size(); j++) {
				TATransition t2 = (TATransition)slTransitions.elementAt(j);
				// the label of t1,t2 matches?
				if (matchTransitionLabel(t1,t2) != null)
					tList.add(t2);
			}

// The following is bad implemented: TRIV says that T \ T | U = empty (or
// more generally R|T \ T|U = R \ U), below I don't check continuation of l[x]
//			/* Optimization [TRIV]: before to call the partitioning algorithm we 
//			 * check if we have l[x] \ l[x] | ..., in which case we can immediately
//			 * answer that the result is the null language (i.e. nothing).
//			 * Note that this optimization will work well only if we have 
//			 * normalized the labels before. Without label normalization you can
//			 * have ~[x] \ l[x] = null that is wrong (the correct answer is
//			 * obviously (~\l)[x])
//			 */
// ... I've deleted the code

			// finally call the partitioning algorithm
			processMatchingTransitions(label,t1,tList,s);
		}
	}
	
	/**
	 * Gets a list of source transitions and a list of destination
	 * transition and returns the list of source transitions normalized.
	 * 
	 * @param sTransitions a list of source transitions
	 * @param tList a list of dest transitions
	 * @return a new normalized source transitions' list
	 */
	private Vector normalizeLabels(Vector sTransitions, Vector tList) {
		// diffTags is the list of tag labels in destination transitions
		HashSet diffTags = new HashSet();
		// diffLiterals is the list of the string literals in dest. transitions
		HashSet diffLiterals = new HashSet();
		// now fill diffTags and diffLiterals
		for (int i=0; i< tList.size(); i++) {
			TATransition t = (TATransition)tList.elementAt(i);
			if (t.label.equals(AnyTagName) || t.label.equals(StringTypeName)) 
				; // the label is #String or ~, do nothing
			else if (t.label.startsWith(StrLiteralPrefix) && 
							 t.label.endsWith(StrLiteralSuffix)) 
				// label is a string literal, add to diffLiterals
				diffLiterals.add(t.label);
			else 
				// label is a tag label, add to diffTags
				diffTags.add(t.label);
		}
		// Now copy the source transitions
		Vector srcTransitions = new Vector(sTransitions);
		Vector newTransitions = new Vector();
		// And normalize them 
		for (int i=0; i< srcTransitions.size(); i++) {
			TATransition t1 = (TATransition)srcTransitions.elementAt(i);
			if (t1.label.equals(AnyTagName) || t1.label.equals(StringTypeName)) {
				Iterator it;
				if (t1.label.equals(AnyTagName)) { 
					// the source label is ~, normalize
					t1.normalize(diffTags);
					it = diffTags.iterator();
				}
				else {
					// the source label is #String, normalize
					t1.normalize(diffLiterals);
					it = diffLiterals.iterator();
				}
				while (it.hasNext()) {
					String l = (String)it.next();
					TATransition newT =
					 new TATransition(l,t1.src,t1.dst,t1.content,t1.variables);
					newTransitions.add(newT);
				}
			}
		}
		srcTransitions.addAll(newTransitions);
		return srcTransitions;
	}
	
	/**
	/* <p>Implements the PARTITIONING algorithm or equivalently  the REC 
	 * rule of the old subtyping algorithm you can find in PhD thesis of 
	 * Hosoya. </p>
	 * <p>Also if it resembles to the subset construction and it has less
	 * or more the same role that subset construction algorithm has in the 
	 * standard string automata theory the two algorithms are significatively 
	 * different (see class documentation to know more). </p>
	 * @param t1 the list of transitions that exit from s1 
	 * (where srcState = &lt;s1,sList&gt;) 
	 * @param tList the list of transitions that exit from states in sList
	 * (where srcState = &lt;s1,sList&gt;) 
	 * @param srcState the source difference state to process
	 */
	private void processMatchingTransitions(String l, TATransition t1, 
													Vector tList,TADifferenceState srcState) {
		/* Given n = tList.size() we need to build 2**n new transitions that 
		 * will start from srcState and will go into 2**n distinct (possibly new) 
		 * states, all with "l" as label. */
		int n = tList.size();
		int subsetNum = 1 << n; // subsetNum = 2 ** n
		/* loop through the first 2**n naturals (c stands for counter).
		 * c is used both as counter and as subset specifier of tList:
		 * the first n digits of the counter represent the elements of tList
		 * (1 is in the subset, 0 is not).
		 * For each loop of this cycle we create a new transition and
		 * possibly a new state */
		for (int c=0; c<subsetNum; c++) {
			/* before to create the new transition we need to compute
			 * the current subset of subtrees to differentiate and the 
			 * current subset of states to differentiate. 
			 * The two subsets are one the complement of the other. 
			 * For example if n=3 and c=2 (010 in binary) I take my subtrees 
			 * to differentiate from the first and the third transition in 
			 * tList (because the first and third digits of c are 0), whereas 
			 * my set of states to differentiate is the singleton given by the
			 * destination state of the second transition in tList (because the 
			 * second digit of c is 0).
			 * For example given tList as follows
			 * 		tList[0] = "q3 = a[x] --> q4"
			 *    tList[1] = "q5 = a[y] --> q6"
			 *    tList[2] = "q7 = a[z] --> q8"
			 * if c=5 I take diffTrees = {x,z} and diffStates = {q6}
			 */ 
			HashSet diffTrees = new HashSet(); // set of strings (names of automata)
			HashSet diffStates = new HashSet(); // set of (destination) states
			if (n==0) {
				// diffTrees = null
				// diffStates = null
			}
			else 
				for (int i=0; i<n; i++) {
					int bitmask = 1 << i;
					if ((c & bitmask) != 0) { 
						diffStates.add(((TATransition)tList.elementAt(i)).dst);
					}
					else {
						diffTrees.add(((TATransition)tList.elementAt(i)).content);
					}
				}
			// find the destination state
			boolean t1dstFinal = t1.dst.isFinal();
			boolean slFinal = isDiffFinal(diffStates);
			// don't create transitions that go in null states
			//if (!t1dstFinal || !slFinal) {
				TADifferenceState destState = getState(t1.dst,diffStates);
				// if destination state doesn't exist create it
				if (destState == null) {
					destState = new TADifferenceState(t1.dst,diffStates);
					states.add(destState);
					/* final if the destination in t1 is final but there are not
					 * final destination states in the list of states to subtract */
					if (t1dstFinal && !slFinal) {
						destState.setFinal(true);
						finalStates.add(destState);
					}
				}
				// create the new transition
				/* the name of the automaton that matches the content is
				 * first x second (e.g. "p1xp2") */
				TA contentAutomaton = getContentDiffAutomaton(t1.content,diffTrees);
				String contentName = contentAutomaton.getName();
				// Note that bindings are discarded (I put variables to null)
				TATransition newT = new TATransition(l,srcState,destState,contentName,null);
				newT.diffLabels.addAll(t1.diffLabels); 
				transitions.add(newT);
				srcState.getTransitions().add(newT);
			//}
		}
	}
	
	private TA getContentDiffAutomaton(String a1Name, HashSet alNames) {
		/* Optimizations:
		 * 1) A \ [] = A (if A is null returns null)
		 * 2) null \ [..] = null
		 * 3) A \ [A] = null
		 */
		if (alNames.size() == 0) {
			// A \ [] = A
			TA a1 = symtab.search_automaton(a1Name);
			if (a1 == null)
				XreError.showerror_and_exit("can't find automaton " + a1Name);
			return a1;
		}
		else if (a1Name.equals(nullAutomatonName)) {
			// null \ [..] = null
			return getNullAutomaton();
		}
		else if (alNames.size() == 1 && alNames.contains(a1Name)) {
			// A \ [A]
			return getNullAutomaton();
		}
		else {
			// search the automaton
			String contentName = getDiffName(a1Name,alNames);
			TA contentAutomaton =
				symtab.search_automaton(contentName);
			if (contentAutomaton == null) {
				// if it doesn't exist create it
				TA a1 = symtab.search_automaton(a1Name);
				if (a1 == null)
					XreError.showerror_and_exit("can't find automaton " + a1Name);
				TA a2 = symtab.getUnionAutomaton(alNames,TABuildPhase.FULL_BUILD,debug);
				if (a2 == null)
					XreError.showerror_and_exit("can't build automaton " + 
																			TAUnion.getUnionName(alNames));
				// create the difference automaton
				contentAutomaton = new TADifference(a1,a2,symtab,debug,out);
				contentAutomaton.update(
					TABuildPhase.SIMPLE_BUILD | 
					TABuildPhase.STEP_NO_UNREACH_STATES |
					TABuildPhase.STEP_NO_UNMATCHED_STATES 
					/* I don't put COMPLETE here as product and difference states
					 * are built complete by itself*/
				);
				if (debug) {
					out.println("\nSubautomaton " + contentName + ":");
					contentAutomaton.dump(out);
				}
			}
			else
				// however be sure that unmatched states are removed
				contentAutomaton.update(TABuildPhase.STEP_NO_UNMATCHED_STATES);

			return contentAutomaton;
		}
	}
	
	private TA getNullAutomaton() {
		TA nullAut = symtab.search_automaton(nullAutomatonName);
		if (nullAut == null) {
			nullAut = new TASimple(nullAutomatonName,symtab,debug,out,null);
		}
		return nullAut;
	}
	
	/* search a state and returns the state or null if not found */
	protected TADifferenceState getState(TAState s1, HashSet sList) {
		// Warning! Note that I should use hashing techniques here
		for (int i=0; i<states.size(); i++) {
			TADifferenceState s = (TADifferenceState)states.elementAt(i);
			if (s.getState1() == s1 && s.getStateList().equals(sList))
				return s;
		}
		return null;
	}

	/**
	 * Checks if there is at least one final state in set given set of
	 * states. Note that the given automaton must be complete for this
	 * function work properly.
	 * @param sList a set of states
	 * @return true if there is at least a final state in the given set
	 */
	private boolean isDiffFinal(HashSet sList) {
		// sList is final if at least one of its members is final
		Iterator it = sList.iterator();
		while (it.hasNext()) {
			TAState s2 = (TAState)it.next();
			if (s2.isFinal()) 
				return true;
		}
		return false;
	}

	/* Checks if two transitions match, i.e. if the label is the same.
	 * Contents can be different. Note that this function takes care to
	 * match the special "~" with all the tags but strings and string 
	 * constants with the type String.
	 * If transitions don't match this function returns null, otherwise 
	 * if transitions match it returns the common label (if one of two 
	 * labels is an instance and the other is a special symbol it returns 
	 * the instance). 
	 */
	private String matchTransitionLabel(TATransition t1, TATransition t2) {
		// N.B.: use matchesLabel of TA?
		String[] names = { t1.label, t2.label	};
		boolean[] isTag = {false, false };
		boolean[] isJolly = {false, false };
		
		for (int i=0; i<2; i++) {
			if (names[i].equals(AnyTagName)) { isTag[i] = true; isJolly[i] = true; }
			else if (names[i].startsWith(StrLiteralPrefix) && names[i].endsWith(StrLiteralSuffix)) {
				isTag[i] = false; isJolly[i] = false;
			}
			else if (names[i].equals(StringTypeName)) {
				isTag[i] = false; isJolly[i] = true;
			}
			else { isTag[i] = true; isJolly[i] = false; }
		}
		if (isTag[0] ^ isTag[1]) return null; // ^ = XOR (are different)
		else if (isJolly[0]) {
			if (!t1.diffLabels.contains(names[1])) return names[1];
			else return null;
		} 
		else if (isJolly[1]) {
			if (!t2.diffLabels.contains(names[0])) return names[0];
			else return null;
		} 
		else if (names[0].equals(names[1])) return names[0];
		else return null;
	}
}
