/*
 * TAUnion.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 15-nov-03
 */
package xrel.analyzer;

import java.io.PrintStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import xrel.XreError;
import xrel.parser.SymTable;

/**
 * TODO insert comment here
 * @author Fabrizio Bisi
 */
public class TAUnion extends TA {
	private HashSet aListNames;
	
	public static String getUnionName(HashSet alNames) {
		return alNames.toString();
	}
	
	public TAUnion(HashSet alNames, SymTable st, 
								 boolean dbg, PrintStream psOut) {
		super(getUnionName(alNames),st,dbg,psOut);
		aListNames = new HashSet(alNames);
		//aListNames.addAll(alNames);
	}
	
	/**
	 * 
	 * @see xrel.analyzer.TA#build()
	 */
	public void build() {
		Iterator it = aListNames.iterator();
		while (it.hasNext()) {
			String aName = (String)it.next();
			TA ai = symtab.search_automaton(aName);
			if (ai==null)
				XreError.showerror_and_exit("can't union as " + aName + 
																	  " doesn't exist");
			insert(ai);
		}
	}

	public void union(TA A) {
		if (A != null && !aListNames.contains(A.toString())) {
			aListNames.add(A.toString());
			insert(A);
		}
	}
	
	/**
	 * N.B.: this function add an automaton also if it is already present!
	 * And it doesn't add the name of the automaton to the list of names 
	 * Merges the given automaton with the current one.
	 * Note that unlike merge it clones states and transitions so that the
	 * resulting automaton will have different states.
	 * Note that status information of the states (the isFinal flag and 
	 * the exiting transitions) is not updated, so you'll need to call
	 * the function completeAutomaton() to complete the new automaton. 
	 * Note also that if you join patterns with this function
	 * you could have a not-linear pattern.
	 * 
	 * @param A the automaton to merge with the current
	 */
	private void insert(TA A) {
		// map from old states to new states
		HashMap stateMap = new HashMap();
		// now clone states
		for (int i = 0; i < A.states.size(); i++) {
			TAState s = (TAState)A.states.elementAt(i);
			TAState newS = (TAState)s.clone();
			states.add(newS);
			if (A.initialStates.contains(s))
				initialStates.add(newS);
			if (A.finalStates.contains(s))
				finalStates.add(newS); // I DON'T set the isFinal flag
			stateMap.put(s,newS);
		}
		// convert epsilon transitions 
		for (int i = 0; i < A.epsilonTransitions.size(); i++) {
			TAEpsTransition t = (TAEpsTransition)A.epsilonTransitions.elementAt(i);
			TAState src = (TAState)stateMap.get(t.src);
			TAState dst = (TAState)stateMap.get(t.dst);
			TAEpsTransition newT = new TAEpsTransition(src,dst);
			epsilonTransitions.add(newT);
		}
		// now convert labeled transitions 
		for (int i = 0; i < A.transitions.size(); i++) {
			TATransition t = (TATransition)A.transitions.elementAt(i);
			TAState src = (TAState)stateMap.get(t.src);
			TAState dst = (TAState)stateMap.get(t.dst);
			// variables are referenced, not copied!
			TATransition newT = new TATransition(t.label,src,dst,t.content,t.variables);
			transitions.add(newT);
		}
		// Warning! exiting transitions and isFinal flag of the new states
		// are not set
	}
}
