/*
 * WriteTokensVisitor.java
 * project: XRel
 * @author Fabrizio
 * Created on 12-giu-03
 */
package xrel.parser;

import java.io.PrintStream;

/**
 * <p>This class is a JJTree visitor that writes out the tree in a flat
 * human-readable representation.
 * Flat means not structured, i.e. on a single row. </p>
 * <p>The peculiarity of this class is that it uses the original syntax tree
 * with token annotations to do the work. </p>
 * @author Fabrizio
 */
public class WriteTokensVisitor implements XRelParserVisitor {
	private int indent = 0;
	protected PrintStream out;

	public WriteTokensVisitor(PrintStream o)
	{
	  out = o;
	}

	public Object print(SimpleNode node, Object data) {
		Token t1 = node.getFirstToken();
		Token t = new Token();
		t.next = t1;
		
		SimpleNode n;
		for (int ord = 0; ord < node.jjtGetNumChildren(); ord++) {
			n = (SimpleNode)node.jjtGetChild(ord);
			while (true) {
				t = t.next;
				if (t == n.getFirstToken()) break;
				print(t);
			}
			n.jjtAccept(this, data);
			t = n.getLastToken();
		}
		
		while (t != node.getLastToken()) {
			t = t.next;
			print(t);
		}

		return data;
	}
  
	protected void print(Token t) {
/*
 		Token tt = t.specialToken;
		if (tt != null) {
			while (tt.specialToken != null) tt = tt.specialToken;
			while (tt != null) {
				out.println(addUnicodeEscapes(tt.image) + " ");
				tt = tt.next;
			}
		}
*/
		if (t != null) out.print(addUnicodeEscapes(t.image) + " ");
	}


	private String addUnicodeEscapes(String str) {
		String retval = "";
		char ch;
		for (int i = 0; i < str.length(); i++) {
			ch = str.charAt(i);
			if ((ch < 0x20 || ch > 0x7e) &&
			ch != '\t' && ch != '\n' && ch != '\r' && ch != '\f') {
				String s = "0000" + Integer.toString(ch, 16);
				retval += "\\u" + s.substring(s.length() - 4, s.length());
			} else {
				retval += ch;
			}
		}
		return retval;
	}

	public Object visit(SimpleNode node, Object data) { return print(node, data); }
	public Object visit(ASTProgram node, Object data) { return print(node, data); }
	public Object visit(ASTTypeDecl node, Object data) { return print(node, data); }
	public Object visit(ASTSwitchCase node, Object data) { return print(node, data); }
	public Object visit(ASTCase node, Object data) { return print(node, data); }
	public Object visit(ASTStatementList node, Object data) { return print(node, data); }
	public Object visit(ASTPar node, Object data) { return print(node, data); }
	public Object visit(ASTComma node, Object data) { return print(node, data); }
	public Object visit(ASTImportSt node, Object data) { return print(node,data); }
	public Object visit(ASTExportTp node, Object data) { return print(node,data); }
	public Object visit(ASTOperator node, Object data) { return print(node, data); }
	public Object visit(ASTTag node, Object data) { return print(node, data); }
	public Object visit(ASTEmpty node, Object data) { return print(node, data); }
	public Object visit(ASTString node, Object data) { return print(node, data); }
	public Object visit(ASTStringLiteral node, Object data) { return print(node,data); }
//	public Object visit(ASTLabel node, Object data) { return print(node, data); }
	public Object visit(ASTIdentifier node, Object data) { return print(node, data); }
	public Object visit(ASTAs node, Object data) { return print(node, data); }

}
