/* Builder.java
 * project: XRel
 * @author Fabrizio
 * Created on 17-giu-03
 */
package xrel.xml;

//import java.io.PrintWriter;
//import java.io.FilterOutputStream;
import java.io.FileOutputStream;
import java.io.FileNotFoundException;
import java.io.PrintStream;

import xrel.XreError;

//import xrel.XreError;

/*
import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
*/

/**
 * <p> Permits to build an XML document. </p>
 * <p> Here's a list of methods available:
 * <ul>
 * <li> openDoc(filename, rootname) creates a new document with root
 * "rootname";</li>
 * <li> closeDoc() closes the open document</li>
 * <li> openNode(tagName) opens a new tag with name "tagName"</li>
 * <li> closeNode(tagName) closes a tag with name "tagName"</li>
 * <li> beginNode(tagName) begins a new tag with name "tagName".
 * The difference with openNode() is that beginNode doesn't begin
 * a new line and doesn't male the indentation level higher</li>
 * <li> endNode(tagName) closes a tag with name "tagName". 
 * The difference with closeNode() is that endNode() simply writes the
 * tag termination, while closeNode() also makes the indentation
 * level lower and writes the tag termination at a new line</li>
 * <li> addNode(tagName,content) writes a new tag with name "tagName"
 * on one only row with its content (a string)</li>
 * <li> writeText(content) writes simple text</li>
 * </ul>
 * </p>
 * <p>.. TODO: to complete..</p>
 * @author Fabrizio Bisi
 */
public class Builder {
	/** Stream to which to print. */
	protected PrintStream fOut; 
	/** Number of current indentations. */
	private int indent = 0; 
	/** The name of the root node. */
	private String root; 
	
	/**
	 * Creates a new XML document with the specified name and with a root
	 * tag with name "rt". Specifically it creates the file, writes the header
	 * and opens the root tag.  
	 * @param filename the name of the XML document
	 * @param rt the name of the root node
	 */
	public void openDoc(String filename, String rt) {
		try {
			FileOutputStream file = new FileOutputStream(filename);
			fOut = new PrintStream(file);
		}
		catch (FileNotFoundException e) {
			XreError.showerror_and_exit("can't open file in writing");
		}
			
		fOut.println("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
		fOut.flush();
		root = rt;
		openNode(rt);
	}
	
	/**
	 * Closes the current document.
	 */
	public void closeDoc() {
		closeNode(root);
		// close file?
	}
	
	/**
	 * Opens a new tag and makes the indention level higher.
	 * @param tagName the name of the tag to open
	 */
	public void openNode(String tagName) {
		fOut.println(indentString() + "<" + tagName + ">");
		++indent;
	}
	
	/**
	 * Closes the specified tag and makes the indentation level lower. 
	 * @param tagName the name of the tag to close
	 */
	public void closeNode(String tagName) {
		--indent;
		fOut.println(indentString() + "</" + tagName + ">");
	}
	
	/**
	 * Begins a new tag without increase the indentation level and without 
	 * go to the next line.
	 * @param tagName the name of the tag to open
	 */
	public void beginNode(String tagName) {
		fOut.print(indentString() + "<" + tagName + ">");
	}
	
	/**
	 * Closes the specified tag without decrease the indentation level.
	 * @param tagName the name of the tag to close
	 */
	public void endNode(String tagName) {
		fOut.println("</" + tagName + ">");
	}

	/**
	 * Adds a node with a string content on one row.
	 * @param tagName
	 * @param content
	 */
	public void addNode(String tagName, String content) {
		if (content.length() > 0)
			fOut.println(indentString() +  
				"<" + tagName + ">" + content + "</" + tagName + ">");
		else fOut.println(indentString() +  "<" + tagName + "/>");
	}
	
	/**
	 * Writes simple text.
	 * @param content
	 */
	public void writeText(String content) {
		// unquote!
		fOut.println(indentString() + content.substring(1,content.length()-1));
	}
	
	/**
	 * Returns the stream specfied at the creation of the object.
	 * @return the print stream
	 */
	public PrintStream getPrintStream() {
		return fOut;
	}
	
	/**
	 * Builds the indentation string.
	 * @return the indentation string.
	 */
	private String indentString() {
	  StringBuffer sb = new StringBuffer();
	  for (int i = 0; i < indent; ++i) sb.append(" ");
	  return sb.toString();
	}
}
