/*
 * XmlDoc.java
 * project: XRel
 * @author Fabrizio Bisi
 * Created on 29-lug-03
 */
package xrel.xml;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import xrel.XreError;
import xrel.parser.ASTComma;
import xrel.parser.ASTEmpty;
//import xrel.parser.ASTLabel;
import xrel.parser.ASTStringLiteral;
import xrel.parser.ASTTag;
import xrel.parser.SimpleNode;
import xrel.parser.XRelParserTreeConstants;

/**
 * Traverses an XML document with DOM and translates it
 * in an internal format tree.
 * @author Fabrizio Bisi
 */
public class XmlDoc implements XRelParserTreeConstants {
	/** Document information. */
	protected static ParserWrapper.DocumentInfo fDocumentInfo;
	/** Default parser name (Xerces). */
	protected static final String DEFAULT_PARSER_NAME = "xrel.xml.Xerces";
	/** The imported document. */
	protected static Document document = null;
	
	/**
	 * <p>Loads the specified file and parses it with DOM. </p>
	 * @param filename the name to load and parse
	 * @return true if successfully loaded
	 */
	public static boolean loadFrom(String filename) {
		ParserWrapper parser = null;

		// create the parser
		try {
			parser = (ParserWrapper)Class.forName(DEFAULT_PARSER_NAME).newInstance();
		}
		catch (Exception e) {
			System.err.println("Unable to instantiate XML parser (" +
												 DEFAULT_PARSER_NAME + ")");
			return false;
		}
		// parse file
		try {
			document = parser.parse(filename);
			ParserWrapper.DocumentInfo documentInfo = parser.getDocumentInfo();
			fDocumentInfo = documentInfo;

			document.normalize();
			return true;
		}
		catch (SAXParseException e) {
			return false; // ignore
		}
		catch (Exception e) {
			System.err.println("XML parse error occurred: " + e.getMessage());
			Exception se = e;
			if (e instanceof SAXException) {
				se = ((SAXException)e).getException();
			}
			if (se != null)
			  se.printStackTrace(System.err);
			else
			  e.printStackTrace(System.err);
			return false;
		}
	}
	
	/**
	 * <p>Converts the XML document imported from file with loadFrom() 
	 * in an internal format. </p>
	 * @return the internal tree that represents the XML document
	 */
	public static SimpleNode convert() {
		return xmlSubtree(document);
	}
	
	/** 
	 * <p>Traverses the specified node recursively and produces
	 * an equivalent internal subtree.</p>
	 * <p>This function <u>phisically</u> does the translation.</p> 
	 * @param node the DOM node to parse
	 * @return the root SimpleNode  
	 */
	protected static SimpleNode xmlSubtree(Node node) {
		// is there anything to do?
		if (node == null) {
			return null;
		}

		int type = node.getNodeType();
		switch (type) {
			case Node.DOCUMENT_NODE:  
				Document document = (Document)node;
				return xmlSubtree(document.getDocumentElement());
				//break;
			
			case Node.ELEMENT_NODE: 
				ASTTag itag = new ASTTag(JJTTAG);
				itag.setName(node.getNodeName());
        
				// there are tags and text sections (maybe mixed)
				SimpleNode icontent = new ASTComma(JJTCOMMA);
				Node child = node.getFirstChild();
				int pos = 0;
				while (child != null) {
					SimpleNode ichild = xmlSubtree(child);
					if (ichild != null)
						icontent.jjtmAddChildExt(ichild,pos++);
					child = child.getNextSibling();
				}
				
				if (icontent.jjtGetNumChildren() == 0) {
					icontent = new ASTEmpty(JJTEMPTY);
				}
				else if (icontent.jjtGetNumChildren() == 1) {
					icontent = (SimpleNode)icontent.jjtGetChild(0);
				}
				else {
				}
				// however I link the content to the tag
				itag.jjtmAddChildExt(icontent,0);
				return itag;
				//break;
			
			case Node.ENTITY_REFERENCE_NODE: 
				/*
				Node child = node.getFirstChild();
				while (child != null) {
					convertTree(child);
					child = child.getNextSibling();
				}
				*/
				return null; //unhandled
				// break;
			
			case Node.CDATA_SECTION_NODE: 
				// sections <![CDATA[]]>
				return null; // unhandled
				//break;
			
			case Node.TEXT_NODE: 
				if (fDocumentInfo != null) {
					Text text = (Text)node;
					String TextStr = text.getData().trim();
					SimpleNode inode;
					if (TextStr.length() > 0) {  
						inode = new ASTStringLiteral(JJTSTRINGLITERAL);
						((ASTStringLiteral)inode).setValue("\"" + TextStr + "\"");
					}
					else {
						inode = null;
					}
					return inode;
				}
				else {
					XreError.showbug_and_exit("What's happening here?");
					return null;
				}
				//break;
			
			default:
				// unhandled (e.g. ATTRIBUTE_NODE)
				//XreError.showbug_and_exit("Unknown XML node");
				return null;
		}
	} 
}
